"""
Random helper module providing secure and non-secure random number generation utilities.
This module offers functions for generating random integers, strings, booleans, and more,
with options for both cryptographically secure and non-secure random generation.

The module uses Python's built-in `secrets` module for secure generation and `random` module
for non-secure generation. All methods support both secure and non-secure modes via the
`secure` parameter.

Copyright (c) 2022 Jim Schilling.

Please keep this header.

This module is licensed under the MIT License.
"""

import random
import string
import sys
from datetime import date, datetime, timedelta
from secrets import randbits
from typing import List, Optional


class RandomHelper:
    """
    A utility class for generating various types of random values.

    This class provides methods for generating random integers, strings, booleans, and more.
    All methods support both cryptographically secure (using secrets module) and non-secure
    (using random module) generation modes.

    Attributes:
        INT64_MAX (int): Maximum value for 64-bit signed integer (2^63 - 1)
        INT64_MIN (int): Minimum value for 64-bit signed integer (-2^63)
        INT64_MASK (int): Bit mask for 64-bit integers (0x7fff_ffff_ffff_ffff)
        ALPHA_CHARS (str): All ASCII letters (a-z, A-Z)
        DIGITS (str): All decimal digits (0-9)
        ALPHANUMERIC_CHARS (str): Combination of letters and digits
        BASE58_CHARS (str): Base58 character set (excluding 0, O, I, l)
    """

    INT64_MAX: int = 2**63 - 1
    INT64_MIN: int = -(2**63)
    INT64_MASK: int = 0x7FFF_FFFF_FFFF_FFFF
    ALPHA_CHARS: str = f"{string.ascii_lowercase}{string.ascii_uppercase}"
    DIGITS: str = "0123456789"
    ALPHANUMERIC_CHARS: str = f"{ALPHA_CHARS}{DIGITS}"
    BASE58_CHARS: str = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"

    @staticmethod
    def as_bytes(
        size: int,
        *,
        secure: Optional[bool] = False
    ) -> bytes:
        """
        Generate random bytes of specified size.

        Args:
            size (int): Number of bytes to generate
            secure (bool, optional): If True, uses secrets.randbits() for cryptographically
                secure generation. If False, uses random.randbytes(). Defaults to False.

        Returns:
            bytes: Random bytes of specified size

        Example:
            >>> RandomHelper.as_bytes(4)
            b'\x12\x34\x56\x78'
            >>> RandomHelper.as_bytes(4, secure=True)  # Cryptographically secure
            b'\x9a\xb2\xc3\xd4'
        """
        if secure:
            bits: int = randbits(size * 8)
            return bits.to_bytes(size, sys.byteorder)
        return random.randbytes(size)

    @classmethod
    def as_int(
        cls,
        size: int = 8,
        *,
        secure: Optional[bool] = False
    ) -> int:
        """
        Generate a random 64-bit integer.

        Args:
            size (int, optional): Number of bytes to generate. Defaults to 8.
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            int: Random 64-bit integer between 0 and 2^63-1

        Example:
            >>> RandomHelper.as_int()
            1234567890123456789
            >>> RandomHelper.as_int(secure=True)  # Cryptographically secure
            9876543210987654321
        """
        return (
            int.from_bytes(cls.as_bytes(size, secure=secure), sys.byteorder)
            & cls.INT64_MAX
        )

    @classmethod
    def as_int_range(
        cls,
        lower: int,
        upper: int,
        *,
        secure: Optional[bool] = False
    ) -> int:
        """
        Generate a random 64-bit integer within a specified range.

        Args:
            lower (int): Lower bound (inclusive)
            upper (int): Upper bound (inclusive)
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            int: Random 64-bit integer between lower and upper (inclusive)

        Raises:
            ValueError: If lower >= upper or range is outside valid 64-bit bounds

        Example:
            >>> random_int_range(1000000, 2000000)
            1567890
            >>> random_int_range(1000000, 2000000, secure=True)  # Cryptographically secure
            1789012
        """
        if lower >= upper:
            raise ValueError("lower must be < upper")
        if lower < cls.INT64_MIN or upper > cls.INT64_MAX:
            raise ValueError("allowed range is -2**63+1 to 2**63-1")
        return int(cls.as_int(secure=secure) % (upper - lower + 1)) + lower

    @classmethod
    def as_float_range(
        cls,
        lower: float,
        upper: float
    ) -> float:
        """
        Generate a random float within a specified range.

        Args:
            lower (float): Lower bound (inclusive)
            upper (float): Upper bound (inclusive)

        Returns:
            float: Random float between lower and upper (inclusive)

        Raises:
            ValueError: If lower >= upper

        Example:
            >>> random_float_range(0.0, 1.0)
            0.56789
            >>> random_float_range(-1.0, 1.0)
            0.12345
        """
        if lower >= upper:
            raise ValueError("lower must be < upper")
        return random.uniform(lower, upper)

    @classmethod
    def as_string(
        cls,
        length: int,
        allowable_chars: str,
        *,
        secure: Optional[bool] = False
    ) -> str:
        """
        Generate a random string of specified length using given characters.

        Args:
            length (int): Length of the string to generate
            allowable_chars (str): Characters to use in the random string
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            str: Random string of specified length

        Raises:
            ValueError: If length < 1 or allowable_chars is empty

        Example:
            >>> random_string(5, "abc")
            'abcba'
            >>> random_string(10, RandomHelperConstants.ALPHANUMERIC_CHARS)
            'aB3cD4eF5g'
        """
        if length < 1:
            raise ValueError("length must be > 0")
        if not allowable_chars:
            raise ValueError("allowable_chars must be at least 1 character")

        return "".join(
            [
                allowable_chars[
                    cls.as_int_range(0, len(allowable_chars) - 1, secure=secure)
                ]
                for _ in range(0, length)
            ]
        )

    @classmethod
    def as_variable_string(
        cls,
        lower: int,
        upper: int,
        allowable_chars: str,
        *,
        secure: Optional[bool] = False
    ) -> str:
        """
        Generate a random string with length between lower and upper bounds.

        Args:
            lower (int): Minimum length (inclusive)
            upper (int): Maximum length (inclusive)
            allowable_chars (str): Characters to use in the random string
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            str: Random string with length between lower and upper

        Raises:
            ValueError: If lower < 0 or lower >= upper

        Example:
            >>> random_variable_string(3, 5, "abc")
            'abcba'
            >>> random_variable_string(5, 10, RandomHelperConstants.ALPHANUMERIC_CHARS)
            'aB3cD4eF'
        """
        if lower < 0 or lower >= upper:
            raise ValueError("lower must be >= 0 and < upper")

        length: int = cls.as_int_range(lower, upper, secure=secure)

        return (
            cls.as_string(length, allowable_chars, secure=secure) if length > 0 else ""
        )

    @classmethod
    def as_alpha(
        cls,
        length: int,
        *,
        secure: Optional[bool] = False
    ) -> str:
        """
        Generate a random string of letters.

        Args:
            length (int): Length of the string to generate
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            str: Random string of letters

        Example:
            >>> random_alpha(5)
            'aBcDe'
            >>> random_alpha(5, secure=True)  # Cryptographically secure
            'XyZab'
        """
        return cls.as_string(length, cls.ALPHA_CHARS, secure=secure)

    @classmethod
    def as_alphanumeric(
        cls,
        length: int,
        *,
        secure: Optional[bool] = False
    ) -> str:
        """
        Generate a random alphanumeric string.

        Args:
            length (int): Length of the string to generate
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            str: Random alphanumeric string

        Example:
            >>> random_alphanumeric(5)
            'aB3cD'
            >>> random_alphanumeric(5, secure=True)  # Cryptographically secure
            'Xy4Za'
        """
        return cls.as_string(length, cls.ALPHANUMERIC_CHARS, secure=secure)

    @classmethod
    def as_numeric(
        cls,
        length: int,
        *,
        secure: Optional[bool] = False
    ) -> str:
        """
        Generate a random numeric string.

        Args:
            length (int): Length of the string to generate
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            str: Random numeric string

        Example:
            >>> random_numeric(5)
            '12345'
            >>> random_numeric(5, secure=True)  # Cryptographically secure
            '98765'
        """
        return cls.as_string(length, cls.DIGITS, secure=secure)

    @classmethod
    def as_base58(
        cls,
        length: int,
        *,
        secure: Optional[bool] = False
    ) -> str:
        """
        Generate a random Base58 string.

        Args:
            length (int): Length of the string to generate
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            str: Random Base58 string

        Example:
            >>> random_base58(5)
            '1aB2c'
            >>> random_base58(5, secure=True)  # Cryptographically secure
            '3xY4z'
        """
        return cls.as_string(length, cls.BASE58_CHARS, secure=secure)

    @classmethod
    def as_variable_base58(
        cls,
        lower: int,
        upper: int,
        *,
        secure: Optional[bool] = False
    ) -> str:
        """
        Generate a random Base58 string with variable length.

        Args:
            lower (int): Minimum length (inclusive)
            upper (int): Maximum length (inclusive)
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            str: Random Base58 string with length between lower and upper

        Example:
            >>> random_variable_base58(3, 5)
            '1aB2'
            >>> random_variable_base58(3, 5, secure=True)  # Cryptographically secure
            '3xY4'
        """
        return cls.as_variable_string(lower, upper, cls.BASE58_CHARS, secure=secure)

    @classmethod
    def as_variable_alpha(
        cls,
        lower: int,
        upper: int,
        *,
        secure: Optional[bool] = False
    ) -> str:
        """
        Generate a random alphabetic string with variable length.

        Args:
            lower (int): Minimum length (inclusive)
            upper (int): Maximum length (inclusive)
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            str: Random alphabetic string with length between lower and upper

        Example:
            >>> random_variable_alpha(3, 5)
            'aBc'
            >>> random_variable_alpha(3, 5, secure=True)  # Cryptographically secure
            'XyZ'
        """
        return cls.as_variable_string(lower, upper, cls.ALPHA_CHARS, secure=secure)

    @classmethod
    def as_variable_alphanumeric(
        cls,
        lower: int,
        upper: int,
        *,
        secure: Optional[bool] = False
    ) -> str:
        """
        Generate a random alphanumeric string with variable length.

        Args:
            lower (int): Minimum length (inclusive)
            upper (int): Maximum length (inclusive)
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            str: Random alphanumeric string with length between lower and upper

        Example:
            >>> random_variable_alphanumeric(3, 5)
            'aB3'
            >>> random_variable_alphanumeric(3, 5, secure=True)  # Cryptographically secure
            'Xy4'
        """
        return cls.as_variable_string(
            lower, upper, cls.ALPHANUMERIC_CHARS, secure=secure
        )

    @classmethod
    def as_variable_numeric(
        cls,
        lower: int,
        upper: int,
        *,
        secure: Optional[bool] = False
    ) -> str:
        """
        Generate a random numeric string with variable length.

        Args:
            lower (int): Minimum length (inclusive)
            upper (int): Maximum length (inclusive)
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            str: Random numeric string with length between lower and upper

        Example:
            >>> random_variable_numeric(3, 5)
            '123'
            >>> random_variable_numeric(3, 5, secure=True)  # Cryptographically secure
            '987'
        """
        return cls.as_variable_string(lower, upper, cls.DIGITS, secure=secure)

    @classmethod
    def as_bool(
        cls,
        *,
        secure: Optional[bool] = False
    ) -> bool:
        """
        Generate a random boolean value.

        Args:
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            bool: Random boolean value

        Example:
            >>> random_bool()
            True
            >>> random_bool(secure=True)  # Cryptographically secure
            False
        """
        return cls.as_int_range(0, 1, secure=secure) == 1

    @classmethod
    def as_masked_string(
        cls,
        mask: str,
        *,
        secure: Optional[bool] = False
    ) -> str:
        """
        Generate a random string based on a mask pattern.

        The mask can contain:
        - '#' for random digits (0-9)
        - '@' for random letters (a-z, A-Z)
        - Other characters are preserved as-is

        Args:
            mask (str): Pattern to generate random string from
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            str: Random string following the mask pattern

        Raises:
            ValueError: If mask is empty or contains no mask characters (# or @)

        Example:
            >>> RandomHelper.as_masked_string('###-@@@')
            '123-abc'
            >>> RandomHelper.as_masked_string('###-@@@', secure=True)  # Cryptographically secure
            '456-xyz'
        """
        if not mask or (mask.count("#") == 0 and mask.count("@") == 0):
            raise ValueError("mask must contain at least one mask char # or @")

        digit_count: int = mask.count("#")
        digits: str = cls.as_numeric(digit_count, secure=secure)
        alpha_count: int = mask.count("@")
        alphas: str = cls.as_alpha(alpha_count, secure=False)
        value: str = mask

        if digit_count:
            for digit in digits:
                value = value.replace("#", digit, 1)

        if alpha_count:
            for alpha in alphas:
                value = value.replace("@", alpha, 1)

        return value

    @classmethod
    def as_sequenced_string(
        cls,
        count: int,
        digits: int,
        *,
        start: Optional[int] = 0,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None
    ) -> List[str]:
        """
        Generate a list of sequentially numbered strings.

        Args:
            count (int): Number of strings to generate
            digits (int): Number of digits in the sequence number (zero-padded)
            start (int, optional): Starting number. Defaults to 0.
            prefix (str, optional): Prefix for each string. Defaults to None.
            suffix (str, optional): Suffix for each string. Defaults to None.

        Returns:
            List[str]: List of sequentially numbered strings

        Raises:
            ValueError: If count < 1, digits < 1, start < 0, or sequence would exceed digit capacity

        Example:
            >>> RandomHelper.as_sequenced_string(3, 3, prefix='ID-')
            ['ID-000', 'ID-001', 'ID-002']
            >>> RandomHelper.as_sequenced_string(3, 3, start=100, prefix='ID-', suffix='-END')
            ['ID-100-END', 'ID-101-END', 'ID-102-END']
        """
        if count < 1:
            raise ValueError("count must be >= 1")
        if digits < 1:
            raise ValueError("digits must be >= 1")
        if start < 0:
            raise ValueError("start must be >= 0")

        max_value: int = 10**digits - 1
        if start + count > max_value:
            raise ValueError(
                f"digits not large enough to hold sequence value of {max_value}"
            )

        prefix = prefix if prefix else ""
        suffix = suffix if suffix else ""
        values: List[str] = []

        for sequence in range(start, start + count):
            values.append(f"{prefix}{sequence:0{digits}}{suffix}")

        return values

    @classmethod
    def as_date(
        cls,
        lower_days: int,
        upper_days: int,
        *,
        base_date: Optional[date] = None,
        secure: Optional[bool] = False
    ) -> date:
        """
        Generate a random date between two days.

        Args:
            lower_days (int): Minimum number of days from today
            upper_days (int): Maximum number of days from today
            base_date (date, optional): Base date to use for generation. Defaults to None.
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            datetime.date: Random date between two days

        Example:
            >>> RandomHelper.as_date(0, 30)
            datetime.date(2025, 6, 16)
            >>> RandomHelper.as_date(0, 30, secure=True)  # Cryptographically secure
            datetime.date(2025, 7, 15)
        """
        return (base_date if base_date else date.today()) + timedelta(
            days=cls.as_int_range(lower_days, upper_days, secure=secure)
        )

    @classmethod
    def as_datetime(
        cls,
        lower_days: int,
        upper_days: int,
        *,
        base_date: Optional[datetime] = None,
        secure: Optional[bool] = False
    ) -> datetime:
        """
        Generate a random datetime between two days.

        Args:
            lower_days (int): Minimum number of days from today
            upper_days (int): Maximum number of days from today
            base_date (datetime, optional): Base date to use for generation. Defaults to None.
            secure (bool, optional): If True, uses cryptographically secure random generation.
                Defaults to False.

        Returns:
            datetime: Random datetime between two days, with random time component

        Example:
            >>> RandomHelper.as_datetime(0, 30)
            datetime.datetime(2025, 6, 16, 14, 30, 45)
            >>> RandomHelper.as_datetime(0, 30, secure=True)  # Cryptographically secure
            datetime.datetime(2025, 7, 15, 9, 15, 30)
        """
        base_date = base_date if base_date else datetime.now()

        days: int = cls.as_int_range(lower_days, upper_days, secure=secure)
        result: datetime = base_date + timedelta(days=days)

        hours: int = cls.as_int_range(0, 23, secure=secure)
        minutes: int = cls.as_int_range(0, 59, secure=secure)
        seconds: int = cls.as_int_range(0, 59, secure=secure)
        microseconds: int = cls.as_int_range(0, 999999, secure=secure)

        return result.replace(
            hour=hours, minute=minutes, second=seconds, microsecond=microseconds
        )



"""
Tests for data validation utilities.
"""

import unittest

from splurge_tools.data_validator import DataValidator


class TestDataValidator(unittest.TestCase):
    def setUp(self):
        self.validator = DataValidator()

    def test_required_validator(self):
        # Test required field validation
        self.validator.add_validator("name", DataValidator.required())

        # Valid data
        data = {"name": "John"}
        errors = self.validator.validate(data)
        self.assertEqual(len(errors), 0)

        # Invalid data - missing field
        data = {}
        errors = self.validator.validate(data)
        self.assertIn("name", errors)
        self.assertIn("Field is required", errors["name"])

        # Invalid data - empty value
        data = {"name": ""}
        errors = self.validator.validate(data)
        self.assertIn("name", errors)

    def test_length_validators(self):
        # Test min and max length validation
        self.validator.add_validator("username", DataValidator.min_length(3))
        self.validator.add_validator("username", DataValidator.max_length(10))

        # Valid data
        data = {"username": "john_doe"}
        errors = self.validator.validate(data)
        self.assertEqual(len(errors), 0)

        # Invalid data - too short
        data = {"username": "jo"}
        errors = self.validator.validate(data)
        self.assertIn("username", errors)

        # Invalid data - too long
        data = {"username": "john_doe_smith"}
        errors = self.validator.validate(data)
        self.assertIn("username", errors)

    def test_pattern_validator(self):
        # Test pattern validation
        self.validator.add_validator(
            "phone", DataValidator.pattern(r"^\d{3}-\d{3}-\d{4}$")
        )

        # Valid data
        data = {"phone": "123-456-7890"}
        errors = self.validator.validate(data)
        self.assertEqual(len(errors), 0)

        # Invalid data
        data = {"phone": "1234567890"}
        errors = self.validator.validate(data)
        self.assertIn("phone", errors)

    def test_range_validator(self):
        # Test range validation
        self.validator.add_validator("age", DataValidator.numeric_range(0, 120))

        # Valid data
        data = {"age": 25}
        errors = self.validator.validate(data)
        self.assertEqual(len(errors), 0)

        # Invalid data
        data = {"age": 150}
        errors = self.validator.validate(data)
        self.assertIn("age", errors)


if __name__ == "__main__":
    unittest.main()



"""
This module contains the TypedTabularDataModel class that extends TabularDataModel
for typed data access.

Copyright (c) 2025, Jim Schilling

Please keep the copyright notice in this file and in the source code files.

This module is licensed under the MIT License.
"""

from dataclasses import dataclass
from typing import Any, Generator, Optional

from splurge_tools.tabular_data_model import TabularDataModel
from splurge_tools.type_helper import DataType, String, profile_values


@dataclass
class TypeConfig:
    """
    Configuration for handling empty and none-like values for a specific data type.
    """
    empty_default: Any
    none_default: Any


class TypedTabularDataModel(TabularDataModel):
    """
    Extends TabularDataModel to provide typed data access.
    Values are converted to native Python types based on inferred column type.
    """

    def __init__(
        self,
        data: list[list[str]],
        *,
        header_rows: int = 1,
        multi_row_headers: int = 1,
        skip_empty_rows: bool = True,
        type_configs: Optional[dict[DataType, TypeConfig]] = None
    ) -> None:
        super().__init__(
            data,
            header_rows=header_rows,
            multi_row_headers=multi_row_headers,
            skip_empty_rows=skip_empty_rows
        )
        self._typed_data = []
        self._type_configs = {
            DataType.BOOLEAN: TypeConfig(False, False),
            DataType.INTEGER: TypeConfig(0, 0),
            DataType.FLOAT: TypeConfig(0.0, 0.0),
            DataType.DATE: TypeConfig(None, None),
            DataType.DATETIME: TypeConfig(None, None),
            DataType.STRING: TypeConfig("", ""),
            DataType.MIXED: TypeConfig("", None),
            DataType.EMPTY: TypeConfig("", ""),
            DataType.NONE: TypeConfig(None, None),
            DataType.TIME: TypeConfig(None, None),
        }
        if type_configs:
            self._type_configs.update(type_configs)
        self._convert_data()

    def _convert_data(self) -> None:
        """
        Convert all data to native Python types based on column types.
        """
        for col_name in self._column_names:
            self.column_type(col_name)
        self._typed_data = [
            [self._convert_value(value, self._column_types[self._column_names[i]]) for i, value in enumerate(row)]
            for row in self._data
        ]

    def _convert_value(
        self,
        value: str,
        data_type: DataType
    ) -> Any:
        """
        Convert a string value to its native Python type based on the DataType.
        """
        type_config: TypeConfig = self._type_configs[data_type]
        if data_type == DataType.MIXED:
            if String.is_none_like(value):
                return type_config.none_default
            return value
        if String.is_empty_like(value):
            return type_config.empty_default
        if String.is_none_like(value):
            return type_config.none_default
        if data_type == DataType.BOOLEAN:
            return String.to_bool(value, default=type_config.empty_default)
        if data_type == DataType.INTEGER:
            return String.to_int(value, default=type_config.empty_default)
        if data_type == DataType.FLOAT:
            return String.to_float(value, default=type_config.empty_default)
        if data_type == DataType.DATE:
            return String.to_date(value, default=type_config.empty_default)
        if data_type == DataType.DATETIME:
            return String.to_datetime(value, default=type_config.empty_default)
        if data_type == DataType.TIME:
            return String.to_time(value, default=type_config.empty_default)
        return value

    def column_values(
        self,
        name: str
    ) -> list[Any]:
        """
        Get all values for a column in their native Python type.
        """
        if name not in self._column_index_map:
            raise ValueError(f"Column name {name} not found")
        col_idx: int = self._column_index_map[name]
        return [row[col_idx] for row in self._typed_data]

    def cell_value(
        self,
        name: str,
        row_index: int
    ) -> Any:
        """
        Get a cell value by column name and row index in its native Python type.
        """
        if name not in self._column_index_map:
            raise ValueError(f"Column name {name} not found")
        if row_index < 0 or row_index >= self._rows:
            raise ValueError(f"Row index {row_index} out of range")
        col_idx: int = self._column_index_map[name]
        return self._typed_data[row_index][col_idx]

    def iter_rows(self) -> Generator[dict[str, Any], None, None]:
        """
        Iterate over rows as dictionaries with native Python types.
        """
        for row in self._typed_data:
            yield dict(zip(self._column_names, row))

    def iter_rows_as_tuples(self) -> Generator[tuple[Any, ...], None, None]:
        """
        Iterate over rows as tuples with native Python types.
        """
        for row in self._typed_data:
            yield tuple(row)

    def row(
        self,
        index: int
    ) -> dict[str, Any]:
        """
        Get a row as a dictionary with native Python types.
        """
        if index < 0 or index >= self._rows:
            raise ValueError(f"Row index {index} out of range")
        return {
            self._column_names[i]: self._typed_data[index][i]
            for i in range(self._columns)
        }

    def row_as_list(
        self,
        index: int
    ) -> list[Any]:
        """
        Get a row as a list with native Python types.
        """
        if index < 0 or index >= self._rows:
            raise ValueError(f"Row index {index} out of range")
        return self._typed_data[index]

    def row_as_tuple(
        self,
        index: int
    ) -> tuple[Any, ...]:
        """
        Get a row as a tuple with native Python types.
        """
        if index < 0 or index >= self._rows:
            raise ValueError(f"Row index {index} out of range")
        return tuple(self._typed_data[index])

    def column_type(
        self,
        name: str
    ) -> DataType:
        """
        Get the inferred data type for a column. Cached for the column.
        """
        if name not in self._column_index_map:
            raise ValueError(f"Column name {name} not found")
        if name not in self._column_types:
            col_idx: int = self._column_index_map[name]
            values: list[str] = [row[col_idx] for row in self._data]
            non_empty_values: list[str] = [
                v for v in values if not String.is_empty_like(v) and not String.is_none_like(v)
            ]
            if non_empty_values:
                inferred_type: DataType = profile_values(non_empty_values)
                if inferred_type != DataType.MIXED:
                    self._column_types[name] = inferred_type
                    return inferred_type
            self._column_types[name] = profile_values(values)
        return self._column_types[name]



import unittest

from splurge_tools.case_helper import CaseHelper


class TestCaseHelper(unittest.TestCase):
    """Test cases for CaseHelper class."""

    def test_to_train(self):
        """Test train case conversion."""
        test_cases = [
            ("hello world", "Hello-World"),
            ("HELLO WORLD", "Hello-World"),
            ("hello-world", "Hello-World"),
            ("hello_world", "Hello-World"),
            ("", ""),
        ]
        for input_str, expected in test_cases:
            with self.subTest(input_str=input_str):
                self.assertEqual(CaseHelper.to_train(input_str), expected)

    def test_to_sentence(self):
        """Test sentence case conversion."""
        test_cases = [
            ("hello world", "Hello world"),
            ("HELLO WORLD", "Hello world"),
            ("hello-world", "Hello world"),
            ("hello_world", "Hello world"),
            ("", ""),
        ]
        for input_str, expected in test_cases:
            with self.subTest(input_str=input_str):
                self.assertEqual(CaseHelper.to_sentence(input_str), expected)

    def test_to_camel(self):
        """Test camel case conversion."""
        test_cases = [
            ("hello world", "helloWorld"),
            ("HELLO WORLD", "helloWorld"),
            ("hello-world", "helloWorld"),
            ("hello_world", "helloWorld"),
            ("", ""),
        ]
        for input_str, expected in test_cases:
            with self.subTest(input_str=input_str):
                self.assertEqual(CaseHelper.to_camel(input_str), expected)

    def test_to_snake(self):
        """Test snake case conversion."""
        test_cases = [
            ("hello world", "hello_world"),
            ("HELLO WORLD", "hello_world"),
            ("hello-world", "hello_world"),
            ("HelloWorld", "helloworld"),
            ("", ""),
        ]
        for input_str, expected in test_cases:
            with self.subTest(input_str=input_str):
                self.assertEqual(CaseHelper.to_snake(input_str), expected)

    def test_to_kebab(self):
        """Test kebab case conversion."""
        test_cases = [
            ("hello world", "hello-world"),
            ("HELLO WORLD", "hello-world"),
            ("hello_world", "hello-world"),
            ("HelloWorld", "helloworld"),
            ("", ""),
        ]
        for input_str, expected in test_cases:
            with self.subTest(input_str=input_str):
                self.assertEqual(CaseHelper.to_kebab(input_str), expected)

    def test_to_pascal(self):
        """Test pascal case conversion."""
        test_cases = [
            ("hello world", "HelloWorld"),
            ("HELLO WORLD", "HelloWorld"),
            ("hello-world", "HelloWorld"),
            ("hello_world", "HelloWorld"),
            ("", ""),
        ]
        for input_str, expected in test_cases:
            with self.subTest(input_str=input_str):
                self.assertEqual(CaseHelper.to_pascal(input_str), expected)


if __name__ == "__main__":
    unittest.main()



"""
Tabular data model classes for structured data operations.

Copyright (c) 2025, Jim Schilling

This module is licensed under the MIT License.
"""

import re
from typing import Generator

from splurge_tools.type_helper import DataType, profile_values


class TabularDataModel:
    """
    Tabular data model for structured data.
    """

    def __init__(
        self,
        data: list[list[str]],
        *,
        header_rows: int = 1,
        multi_row_headers: int = 1,
        skip_empty_rows: bool = True
    ) -> None:
        """
        Initialize TabularDataModel.

        Args:
            data (list[list[str]]): Raw data rows.
            header_rows (int): Number of header rows.
            multi_row_headers (int): Number of rows to merge for column names.
            skip_empty_rows (bool): Skip empty rows in data.

        Raises:
            ValueError: If data or header configuration is invalid.
        """
        if data is None or len(data) == 0:
            raise ValueError("Data is required")
        if header_rows < 0:
            raise ValueError("Header rows must be greater than or equal to 0")
        if header_rows > 0 and multi_row_headers > header_rows:
            raise ValueError("Column names span must be less than or equal to header rows")
        if header_rows > 0 and multi_row_headers == 0:
            raise ValueError("Column names span must be greater than 0 if header rows are greater than 0")

        self._raw_data = data
        self._header_rows = header_rows
        self._multi_row_headers = multi_row_headers
        self._header_data = data[:header_rows] if header_rows > 0 else []
        self._data = (
            self._normalize_data_model(data[header_rows:], skip_empty_rows)
            if header_rows > 0
            else self._normalize_data_model(data, skip_empty_rows)
        )
        self._header_columns = len(self._header_data[0]) if len(self._header_data) > 0 else 0
        self._columns = len(self._data[0]) if len(self._data) > 0 else 0
        self._rows = len(self._data) if len(self._data) > 0 else 0

        if header_rows > 1 and multi_row_headers > 1:
            merged_headers: list[str] = []
            for i in range(min(multi_row_headers, len(self._header_data))):
                row = self._header_data[i]
                while len(merged_headers) < len(row):
                    merged_headers.append("")
                for j, name in enumerate(row):
                    if merged_headers[j]:
                        merged_headers[j] = f"{merged_headers[j]}_{name}"
                    else:
                        merged_headers[j] = name
            self._header_data = [merged_headers]

        self._column_names = self._header_data[0] if len(self._header_data) > 0 else []
        self._column_names = [re.sub(r"\s+", " ", name).strip() for name in self._column_names]

        if len(self._column_names) == 0:
            self._column_names = [f"column_{i}" for i in range(self._columns)]

        while len(self._column_names) < self._columns:
            self._column_names.append(f"column_{len(self._column_names)}")

        self._column_names = [name if name else f"column_{i}" for i, name in enumerate(self._column_names)]
        self._column_index_map = {name: i for i, name in enumerate(self._column_names)}
        self._column_types = {}

    @property
    def column_names(self) -> list[str]:
        """
        List of column names.
        """
        return self._column_names

    def column_index(
        self,
        name: str
    ) -> int:
        """
        Get the column index for a given name.

        Args:
            name (str): Column name.

        Returns:
            int: Column index.

        Raises:
            ValueError: If column name is not found.
        """
        if name not in self._column_index_map:
            raise ValueError(f"Column name {name} not found")
        return self._column_index_map[name]

    @property
    def row_count(self) -> int:
        """
        Number of rows.
        """
        return self._rows

    @property
    def column_count(self) -> int:
        """
        Number of columns.
        """
        return self._columns

    def column_type(
        self,
        name: str
    ) -> DataType:
        """
        Get the inferred data type for a column (cached).

        Args:
            name (str): Column name.

        Returns:
            DataType: Inferred data type.

        Raises:
            ValueError: If column name is not found.
        """
        if name not in self._column_index_map:
            raise ValueError(f"Column name {name} not found")
        if name not in self._column_types:
            col_idx: int = self._column_index_map[name]
            values: list[str] = [row[col_idx] for row in self._data]
            self._column_types[name] = profile_values(values)
        return self._column_types[name]

    def column_values(
        self,
        name: str
    ) -> list[str]:
        """
        Get all values for a column.

        Args:
            name (str): Column name.

        Returns:
            list[str]: Values in the column.

        Raises:
            ValueError: If column name is not found.
        """
        if name not in self._column_index_map:
            raise ValueError(f"Column name {name} not found")
        col_idx: int = self._column_index_map[name]
        return [row[col_idx] for row in self._data]

    def cell_value(
        self,
        name: str,
        row_index: int
    ) -> str:
        """
        Get a cell value by column name and row index.

        Args:
            name (str): Column name.
            row_index (int): Row index (0-based).

        Returns:
            str: Cell value.

        Raises:
            ValueError: If column name is not found or row index is out of range.
        """
        if name not in self._column_index_map:
            raise ValueError(f"Column name {name} not found")
        if row_index < 0 or row_index >= self._rows:
            raise ValueError(f"Row index {row_index} out of range")
        col_idx: int = self._column_index_map[name]
        return self._data[row_index][col_idx]

    def __iter__(self) -> Generator[list[str], None, None]:
        return iter(self._data)

    def iter_rows(self) -> Generator[dict[str, str], None, None]:
        """
        Iterate over rows as dictionaries.

        Yields:
            dict[str, str]: Row as a dictionary.
        """
        for row in self._data:
            yield dict(zip(self._column_names, row))

    def iter_rows_as_tuples(self) -> Generator[tuple[str, ...], None, None]:
        """
        Iterate over rows as tuples.

        Yields:
            tuple[str, ...]: Row as a tuple.
        """
        for row in self._data:
            yield tuple(row)

    def row(
        self,
        index: int
    ) -> dict[str, str]:
        """
        Get a row as a dictionary.

        Args:
            index (int): Row index (0-based).

        Returns:
            dict[str, str]: Row as a dictionary.
        """
        return {
            self._column_names[i]: self._data[index][i] for i in range(self._columns)
        }

    def row_as_list(
        self,
        index: int
    ) -> list[str]:
        """
        Get a row as a list.

        Args:
            index (int): Row index (0-based).

        Returns:
            list[str]: Row as a list.
        """
        return self._data[index]

    def row_as_tuple(
        self,
        index: int
    ) -> tuple[str, ...]:
        """
        Get a row as a tuple.

        Args:
            index (int): Row index (0-based).

        Returns:
            tuple[str, ...]: Row as a tuple.
        """
        return tuple(self._data[index])

    @staticmethod
    def _normalize_data_model(
        rows: list[list[str]],
        skip_empty_rows: bool = True
    ) -> list[list[str]]:
        """
        Normalize the data model (pad rows, optionally skip empty rows).

        Args:
            rows (list[list[str]]): Data rows.
            skip_empty_rows (bool): Skip empty rows if True.

        Returns:
            list[list[str]]: Normalized data rows.
        """
        if len(rows) == 0:
            return []
        max_column_count: int = max(len(row) for row in rows)
        normalized_rows: list[list[str]] = []
        for row in rows:
            if len(row) < max_column_count:
                row = row + [""] * (max_column_count - len(row))
            normalized_rows.append(row)
        if skip_empty_rows:
            normalized_rows = [
                row
                for row in normalized_rows
                if not all(cell.strip() == "" for cell in row)
            ]
        return normalized_rows



"""
Data transformation utilities for tabular data models.

Provides pivot, melt, group-by, and column transformation operations for TabularDataModel and TypedTabularDataModel.

Copyright (c) 2024 Jim Schilling

This module is licensed under the MIT License.
"""

from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Union

from splurge_tools.tabular_data_model import TabularDataModel
from splurge_tools.typed_tabular_data_model import TypedTabularDataModel


class DataTransformer:
    """
    Utility for transforming tabular data models.

    Supports pivot, melt, group-by, and column transformation operations.
    """

    def __init__(
        self,
        data_model: Union[TabularDataModel, TypedTabularDataModel]
    ) -> None:
        """
        Initialize the DataTransformer.

        Args:
            data_model (TabularDataModel | TypedTabularDataModel): The data model to transform.
        """
        self._model = data_model

    def pivot(
        self,
        index_cols: List[str],
        columns_col: str,
        values_col: str,
        *,
        agg_func: Optional[Callable[[List[Any]], Any]] = None
    ) -> TabularDataModel:
        """
        Pivot the data model to create a cross-tabulation.

        Args:
            index_cols (List[str]): Columns to use as row indices.
            columns_col (str): Column to use as new columns.
            values_col (str): Column to use as values.
            agg_func (Optional[Callable[[List[Any]], Any]]): Aggregation function for duplicates.

        Returns:
            TabularDataModel: Pivoted data model.

        Raises:
            ValueError: If columns are invalid or duplicates exist without agg_func.
        """
        for col in index_cols + [columns_col, values_col]:
            if col not in self._model.column_names:
                raise ValueError(f"Column {col} not found in data model")

        grouped_data: defaultdict[Any, list[Any]] = defaultdict(list)
        duplicate_keys: set[Any] = set()

        for row in self._model.iter_rows():
            index_key = tuple(row[col] for col in index_cols)
            col_value = row[columns_col]
            val_value = row[values_col]
            if any(existing_col == col_value for existing_col, _ in grouped_data[index_key]):
                duplicate_keys.add(index_key)
            grouped_data[index_key].append((col_value, val_value))

        if duplicate_keys and not agg_func:
            duplicate_examples = list(duplicate_keys)[:3]
            raise ValueError(
                f"Duplicate values found for index keys: {duplicate_examples}. "
                "Please provide an aggregation function to handle duplicates."
            )

        unique_columns = sorted(
            {col for group in grouped_data.values() for col, _ in group}
        )
        header = index_cols + list(unique_columns)

        new_data: list[list[Any]] = []
        for index_key, group in grouped_data.items():
            row_data = list(index_key)
            value_dict = dict(group)
            if agg_func is not None:
                value_dict = {
                    k: agg_func([v for c, v in group if c == k]) for k in unique_columns
                }
            for col in unique_columns:
                row_data.append(str(value_dict.get(col, "")))
            new_data.append(row_data)

        return TabularDataModel([header] + new_data)

    def melt(
        self,
        id_vars: List[str],
        value_vars: List[str],
        *,
        var_name: str = "variable",
        value_name: str = "value"
    ) -> TabularDataModel:
        """
        Unpivot columns into rows (wide to long format).

        Args:
            id_vars (List[str]): Columns to use as identifier variables.
            value_vars (List[str]): Columns to unpivot.
            var_name (str): Name for the variable column.
            value_name (str): Name for the value column.

        Returns:
            TabularDataModel: Melted data model.

        Raises:
            ValueError: If columns are invalid.
        """
        for col in id_vars + value_vars:
            if col not in self._model.column_names:
                raise ValueError(f"Column {col} not found in data model")

        header = id_vars + [var_name, value_name]
        new_data: list[list[Any]] = [
            [*(row[col] for col in id_vars), var, row[var]]
            for row in self._model.iter_rows() for var in value_vars
        ]
        return TabularDataModel([header] + new_data)

    def group_by(
        self,
        group_cols: List[str],
        agg_dict: Dict[str, Callable[[List[Any]], Any]]
    ) -> TabularDataModel:
        """
        Group data by columns and aggregate.

        Args:
            group_cols (List[str]): Columns to group by.
            agg_dict (Dict[str, Callable[[List[Any]], Any]]): Aggregation functions per column.

        Returns:
            TabularDataModel: Grouped and aggregated data model.

        Raises:
            ValueError: If columns are invalid.
        """
        for col in group_cols + list(agg_dict.keys()):
            if col not in self._model.column_names:
                raise ValueError(f"Column {col} not found in data model")

        grouped_data: defaultdict[Any, dict[str, list[Any]]] = defaultdict(lambda: defaultdict(list))
        for row in self._model.iter_rows():
            group_key = tuple(row[col] for col in group_cols)
            for col in agg_dict.keys():
                grouped_data[group_key][col].append(row[col])

        header = group_cols + list(agg_dict.keys())
        new_data: list[list[Any]] = [
            list(group_key) + [str(agg_dict[col](agg_values[col])) for col in agg_dict.keys()]
            for group_key, agg_values in grouped_data.items()
        ]
        return TabularDataModel([header] + new_data)

    def transform_column(
        self,
        column: str,
        transform_func: Callable[[Any], Any]
    ) -> TabularDataModel:
        """
        Transform a column using a function.

        Args:
            column (str): Name of column to transform.
            transform_func (Callable[[Any], Any]): Function to apply to each value.

        Returns:
            TabularDataModel: Data model with transformed column.

        Raises:
            ValueError: If column is invalid.
        """
        if column not in self._model.column_names:
            raise ValueError(f"Column {column} not found in data model")

        new_data: list[list[Any]] = [
            [str(transform_func(row[column])) if col == column else row[col] for col in self._model.column_names]
            for row in self._model.iter_rows()
        ]
        return TabularDataModel([self._model.column_names] + new_data)



"""
Data validation utilities.

This module provides classes for data validation operations.

Copyright (c) 2025, Jim Schilling

This module is licensed under the MIT License.
"""

import re
from typing import Any, Callable, Dict, List, Union


class DataValidator:
    """
    A class for validating data against various rules and constraints.
    """

    def __init__(self) -> None:
        self._validators: Dict[str, List[Callable[[Any], bool]]] = {}
        self._custom_validators: Dict[str, Callable[[Any], bool]] = {}

    def add_validator(
        self,
        field: str,
        validator: Callable[[Any], bool]
    ) -> None:
        """
        Add a validator function for a specific field.

        Args:
            field: The field name to validate
            validator: A function that takes a value and returns True if valid
        """
        if field not in self._validators:
            self._validators[field] = []
        self._validators[field].append(validator)

    def add_custom_validator(
        self,
        name: str,
        validator: Callable[[Any], bool]
    ) -> None:
        """
        Add a named custom validator that can be reused.

        Args:
            name: Unique name for the validator
            validator: A function that takes a value and returns True if valid
        """
        self._custom_validators[name] = validator

    def validate(
        self,
        data: Dict[str, Any]
    ) -> Dict[str, List[str]]:
        """
        Validate all fields in the data dictionary.

        Args:
            data: Dictionary of field names and values to validate

        Returns:
            Dictionary mapping field names to lists of error messages
        """
        errors: Dict[str, List[str]] = {}

        for field, validators in self._validators.items():
            if field not in data:
                errors[field] = ["Field is required"]
                continue

            value = data[field]
            for validator in validators:
                if not validator(value):
                    if field not in errors:
                        errors[field] = []
                    errors[field].append(f"Validation failed for {field}")

        return errors

    @staticmethod
    def required() -> Callable[[Any], bool]:
        """Validator that checks if a value is not None or empty."""
        return lambda x: x is not None and str(x).strip() != ""

    @staticmethod
    def min_length(
        length: int
    ) -> Callable[[Any], bool]:
        """Validator that checks if a string has minimum length."""
        return lambda x: len(str(x)) >= length

    @staticmethod
    def max_length(
        length: int
    ) -> Callable[[Any], bool]:
        """Validator that checks if a string has maximum length."""
        return lambda x: len(str(x)) <= length

    @staticmethod
    def pattern(
        regex: str
    ) -> Callable[[Any], bool]:
        """Validator that checks if a string matches a regex pattern."""
        pattern = re.compile(regex)
        return lambda x: bool(pattern.match(str(x)))

    @staticmethod
    def numeric_range(
        min_val: Union[int, float],
        max_val: Union[int, float]
    ) -> Callable[[Any], bool]:
        """Validator that checks if a number is within a range."""
        return lambda x: min_val <= float(x) <= max_val



"""
Text file utility functions for common file operations.

This module provides helper methods for working with text files, including
line counting, file previewing, and file loading capabilities. The TextFileHelper
class implements static methods for efficient file operations without requiring
class instantiation.

Key features:
- Line counting for text files
- File previewing with configurable line limits
- Complete file loading with header/footer skipping
- Configurable whitespace handling and encoding

Copyright (c) 2025 Jim Schilling

Please preserve this header and all related material when sharing!

This software is licensed under the MIT License.
"""

from os import PathLike
from typing import List, Union


class TextFileHelper:
    """
    Utility class for text file operations.
    All methods are static and memory efficient.
    """

    @staticmethod
    def line_count(
        file_name: Union[PathLike, str],
        *,
        encoding: str = "utf-8"
    ) -> int:
        """
        Count the number of lines in a text file.

        This method efficiently counts lines by iterating through the file
        without loading it entirely into memory.

        Args:
            file_name: Path to the text file
            encoding: File encoding to use (default: 'utf-8')

        Returns:
            int: Number of lines in the file

        Raises:
            FileNotFoundError: If the specified file doesn't exist
            IOError: If there are issues reading the file
            UnicodeDecodeError: If the file cannot be decoded with the specified encoding
        """
        with open(file_name, "r", encoding=encoding) as stream:
            return sum(1 for _ in stream)

    @staticmethod
    def preview(
        file_name: Union[PathLike, str],
        *,
        max_lines: int = 100,
        strip: bool = True,
        encoding: str = "utf-8"
    ) -> List[str]:
        """
        Preview the first N lines of a text file.

        This method reads up to max_lines from the beginning of the file,
        optionally stripping whitespace from each line.

        Args:
            file_name: Path to the text file
            max_lines: Maximum number of lines to read (default: 100)
            strip: Whether to strip whitespace from lines (default: True)
            encoding: File encoding to use (default: 'utf-8')

        Returns:
            List[str]: List of lines from the file

        Raises:
            ValueError: If max_lines < 1
            FileNotFoundError: If the specified file doesn't exist
            IOError: If there are issues reading the file
            UnicodeDecodeError: If the file cannot be decoded with the specified encoding
        """
        if max_lines < 1:
            raise ValueError("TextFileHelper.preview: max_lines is less than 1")
        lines: List[str] = []
        with open(file_name, "r", encoding=encoding) as stream:
            for _ in range(max_lines):
                line = stream.readline()
                if not line:
                    break
                lines.append(line.strip() if strip else line.rstrip("\n"))
        return lines

    @staticmethod
    def load(
        file_name: Union[PathLike, str],
        *,
        strip: bool = True,
        encoding: str = "utf-8",
        skip_header_rows: int = 0,
        skip_footer_rows: int = 0
    ) -> List[str]:
        """
        Load the entire contents of a text file into a list of strings.

        This method reads the complete file into memory, with options to
        strip whitespace from each line and skip header/footer rows.

        Args:
            file_name: Path to the text file
            strip: Whether to strip whitespace from lines (default: True)
            encoding: File encoding to use (default: 'utf-8')
            skip_header_rows: Number of rows to skip from the start (default: 0)
            skip_footer_rows: Number of rows to skip from the end (default: 0)

        Returns:
            List[str]: List of all lines from the file, excluding skipped rows

        Raises:
            FileNotFoundError: If the specified file doesn't exist
            IOError: If there are issues reading the file
            UnicodeDecodeError: If the file cannot be decoded with the specified encoding
        """
        skip_header_rows = max(0, skip_header_rows)
        skip_footer_rows = max(0, skip_footer_rows)
        with open(file_name, "r", encoding=encoding) as stream:
            for _ in range(skip_header_rows):
                if not stream.readline():
                    return []
            lines: List[str] = [line.strip() if strip else line.rstrip("\n") for line in stream]
            if skip_footer_rows > 0:
                if skip_footer_rows >= len(lines):
                    return []
                lines = lines[:-skip_footer_rows]
            return lines



"""Unit tests for DSVHelper class."""

import unittest

from splurge_tools.tabular_data_model import TabularDataModel
from splurge_tools.type_helper import DataType


class TestTabularDataModel(unittest.TestCase):
    """Test cases for TabularDataModel class."""

    def setUp(self):
        """Set up test data."""
        self.sample_data = [
            ["Name", "Age", "City"],  # Header
            ["John", "30", "New York"],
            ["Jane", "25", "Boston"],
            ["Bob", "35", "Chicago"],
        ]

        # Data with mixed types
        self.mixed_type_data = [
            ["Name", "Age", "Score", "IsActive", "Date"],
            ["John", "30", "95.5", "true", "2024-01-01"],
            ["Jane", "25", "88.0", "false", "2024-01-02"],
            ["Bob", "35", "92.5", "true", "2024-01-03"],
        ]

    def test_basic_initialization(self):
        """Test basic model initialization."""
        model = TabularDataModel(self.sample_data)
        self.assertEqual(model.column_names, ["Name", "Age", "City"])
        self.assertEqual(model.row_count, 3)
        self.assertEqual(model.column_count, 3)

    def test_column_index(self):
        """Test column index mapping."""
        model = TabularDataModel(self.sample_data)
        self.assertEqual(model.column_index("Name"), 0)
        self.assertEqual(model.column_index("Age"), 1)
        self.assertEqual(model.column_index("City"), 2)

    def test_row_access(self):
        """Test row access methods."""
        model = TabularDataModel(self.sample_data)

        # Test row as dictionary
        row_dict = model.row(0)
        self.assertEqual(row_dict, {"Name": "John", "Age": "30", "City": "New York"})

        # Test row as list
        row_list = model.row_as_list(1)
        self.assertEqual(row_list, ["Jane", "25", "Boston"])

        # Test row as tuple
        row_tuple = model.row_as_tuple(2)
        self.assertEqual(row_tuple, ("Bob", "35", "Chicago"))

    def test_iteration(self):
        """Test model iteration."""
        model = TabularDataModel(self.sample_data)
        rows = list(model)
        self.assertEqual(len(rows), 3)
        self.assertEqual(rows[0], ["John", "30", "New York"])

    def test_data_normalization(self):
        """Test data normalization with uneven rows."""
        data = [
            ["Name", "Age", "City"],
            ["John", "30"],
            ["Jane", "25", "Boston", "Extra"],
            ["Bob"],
        ]
        model = TabularDataModel(data)
        self.assertEqual(model.column_count, 4)
        self.assertEqual(
            model.row(0), {"Name": "John", "Age": "30", "City": "", "column_3": ""}
        )
        self.assertEqual(
            model.row(1),
            {"Name": "Jane", "Age": "25", "City": "Boston", "column_3": "Extra"},
        )
        self.assertEqual(
            model.row(2), {"Name": "Bob", "Age": "", "City": "", "column_3": ""}
        )

    def test_skip_empty_rows(self):
        """Test skipping empty rows."""
        data = [
            ["Name", "Age", "City"],
            ["John", "30", "New York"],
            ["", "", ""],
            ["Jane", "25", "Boston"],
        ]
        model = TabularDataModel(data, skip_empty_rows=True)
        self.assertEqual(model.row_count, 2)

        model = TabularDataModel(data, skip_empty_rows=False)
        self.assertEqual(model.row_count, 3)

    def test_no_header(self):
        """Test model without header."""
        data = [["John", "30", "New York"], ["Jane", "25", "Boston"]]
        model = TabularDataModel(data, header_rows=0)
        self.assertEqual(model.column_names, ["column_0", "column_1", "column_2"])
        self.assertEqual(model.row_count, 2)
        self.assertEqual(
            model.row(0), {"column_0": "John", "column_1": "30", "column_2": "New York"}
        )
        self.assertEqual(
            model.row(1), {"column_0": "Jane", "column_1": "25", "column_2": "Boston"}
        )

    def test_invalid_initialization(self):
        """Test invalid initialization parameters."""
        with self.assertRaises(ValueError):
            TabularDataModel(None)

        with self.assertRaises(ValueError):
            TabularDataModel([])

        with self.assertRaises(ValueError):
            TabularDataModel(self.sample_data, header_rows=-1)

        with self.assertRaises(ValueError):
            TabularDataModel(self.sample_data, header_rows=1, multi_row_headers=2)

    def test_multi_row_headers(self):
        """Test multi-row header handling."""
        data = [
            ["Category", "Details", "Location"],  # First header row
            ["Name", "Age", "City"],  # Second header row
            ["John", "30", "New York"],
            ["Jane", "25", "Boston"],
        ]
        model = TabularDataModel(data, header_rows=2, multi_row_headers=2)
        self.assertEqual(
            model.column_names, ["Category_Name", "Details_Age", "Location_City"]
        )
        self.assertEqual(model.row_count, 2)
        self.assertEqual(
            model.row(0),
            {"Category_Name": "John", "Details_Age": "30", "Location_City": "New York"},
        )

    def test_iter_rows(self):
        """Test iteration over rows as dictionaries."""
        model = TabularDataModel(self.sample_data)
        rows = list(model.iter_rows())
        self.assertEqual(len(rows), 3)
        self.assertEqual(rows[0], {"Name": "John", "Age": "30", "City": "New York"})
        self.assertEqual(rows[1], {"Name": "Jane", "Age": "25", "City": "Boston"})
        self.assertEqual(rows[2], {"Name": "Bob", "Age": "35", "City": "Chicago"})

    def test_iter_rows_as_tuples(self):
        """Test iteration over rows as tuples."""
        model = TabularDataModel(self.sample_data)
        rows = list(model.iter_rows_as_tuples())
        self.assertEqual(len(rows), 3)
        self.assertEqual(rows[0], ("John", "30", "New York"))
        self.assertEqual(rows[1], ("Jane", "25", "Boston"))
        self.assertEqual(rows[2], ("Bob", "35", "Chicago"))

    def test_column_type(self):
        """Test column type inference."""
        model = TabularDataModel(self.mixed_type_data)

        # Test string column
        self.assertEqual(model.column_type("Name"), DataType.STRING)

        # Test integer column
        self.assertEqual(model.column_type("Age"), DataType.INTEGER)

        # Test float column
        self.assertEqual(model.column_type("Score"), DataType.FLOAT)

        # Test boolean column
        self.assertEqual(model.column_type("IsActive"), DataType.BOOLEAN)

        # Test date column
        self.assertEqual(model.column_type("Date"), DataType.DATE)

        # Test invalid column name
        with self.assertRaises(ValueError):
            model.column_type("InvalidColumn")

    def test_column_values(self):
        """Test getting column values."""
        model = TabularDataModel(self.sample_data)

        # Test valid column
        self.assertEqual(model.column_values("Name"), ["John", "Jane", "Bob"])
        self.assertEqual(model.column_values("Age"), ["30", "25", "35"])
        self.assertEqual(model.column_values("City"), ["New York", "Boston", "Chicago"])

        # Test invalid column name
        with self.assertRaises(ValueError):
            model.column_values("InvalidColumn")

    def test_cell_value(self):
        """Test getting cell values."""
        model = TabularDataModel(self.sample_data)

        # Test valid cells
        self.assertEqual(model.cell_value("Name", 0), "John")
        self.assertEqual(model.cell_value("Age", 1), "25")
        self.assertEqual(model.cell_value("City", 2), "Chicago")

        # Test invalid column name
        with self.assertRaises(ValueError):
            model.cell_value("InvalidColumn", 0)

        # Test invalid row index
        with self.assertRaises(ValueError):
            model.cell_value("Name", -1)
        with self.assertRaises(ValueError):
            model.cell_value("Name", 3)

    def test_column_type_caching(self):
        """Test that column types are cached."""
        model = TabularDataModel(self.mixed_type_data)

        # First call should compute the type
        type1 = model.column_type("Age")

        # Second call should use cached value
        type2 = model.column_type("Age")

        self.assertEqual(type1, type2)
        self.assertEqual(type1, DataType.INTEGER)


if __name__ == "__main__":
    unittest.main()
